#!/usr/bin/env python3
# encoding: utf-8
'''
LMSmusly -- integrates musly into LMS

@author:     R.S.U.

@copyright:  2017 R.S.U. All rights reserved.

@license:    GPL v3

@contact:    https://www.nexus0.net/pub/sw/lmsmusly
@deffield    updated: Updated
'''

import sys, os, logging, time, signal
import sqlite3, atexit

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

from lmsmusly.libmusly import musly
from lmsmusly.lms.server import LMSServer
from lmsmusly.lms.worker import LMSWorker
from lmsmusly.utils import config, tools

__all__ = []
__version__ = '0.1.1'
__date__ = '2017-12-21'
__updated__ = '2018-01-18'

DEBUG = 1
TESTRUN = 0
PROFILE = 0

_LOGGER = logging.getLogger(__name__)


class CLIError(Exception):
	'''Generic exception to raise and log different fatal errors.'''
	def __init__(self, msg):
		super(CLIError).__init__(type(self))
		self.msg = "E: %s" % msg
	def __str__(self):
		return self.msg
	def __unicode__(self):
		return self.msg



def test_musly( mus):
	fpath='/home/luser/media/mp3/new/0misc'
	#fpath = '/home/luser/tmp/sshfs'

	#allfiles = tools.get_files_dir( fpath)
#	allfiles = tools.get_files_inputfile('/home/luser/tmp/bo-slowinstr2.m3u')
#	mtracks = mus.analyze_files( allfiles, "/home/luser/tmp/sshfs:/data/music", True)
	#_LOGGER.info( "analyze_files done: {}".format( len( mtracks)))
	
	#(paths, mtracks) = mus.get_alltracks_db()
#		_LOGGER.debug( "paths: {}".format( repr( paths)))

	#mus.get_track_db( '/home/luser/media/mp3/new/0misc/Portishead - Sour Times (Acapella).mp3')

	(allfiles, mtracks) = mus.get_alltracks_db()
	mtrackids = mus.add_tracks( mtracks)
	mta = musly.MuslyTracksAdded( allfiles, mtracks, mtrackids)
	#_LOGGER.debug( "MuslyTracksAdded: {}".format( mta))
	_LOGGER.debug( "get_numtracks: {}".format( mus.get_numtracks()))
	( rt, rsims) = mus.get_similars( mtracks, mtrackids, 69)
	for i in range( len(rt)):
		_LOGGER.info( "result id: {:3} sim: {:8.6f} file: {}".format( rt[i], rsims[i], allfiles[ rt[i]]))

	_LOGGER.debug( "Stats: {}".format( repr(mus.get_stats())))

def test_lworker( mus, lms):
	lsmworker = LMSWorker( lms, mus, None, None)
	lsmworker.start()
	lsmworker.unpause()
	time.sleep(2)
	lsmworker.pause()
	time.sleep(2)
	lsmworker.unpause()
	time.sleep(2)
	lsmworker.control("exit")
	time.sleep(2)

def test_lms( lms):
	lms.update()
#	_LOGGER.debug( "test_lms: lms= {}".format( lms))
#	_LOGGER.debug( "test_lms: url= {}".format( lms._url))
#	_LOGGER.debug( "test_lms: playlist= {}".format( repr( lms.player( config.LMS_PLAYER).get_playlist())))
#	_LOGGER.debug( "test_lms: curr track= {}".format( repr( lms.player( config.LMS_PLAYER).get_track( lms.player( config.LMS_PLAYER).get_current_index()))))
	_LOGGER.debug( "test_lms: curr track url= {}".format( repr( lms.player( config.LMS_PLAYER).get_track_fileurl( lms.player( config.LMS_PLAYER).get_current_index()))))
#	lms.player( config.LMS_PLAYER).playlist_add('/data/music/Amber Asylum - Frozen in Amber/07 - Aurora.flac')

def test_lmworker( mus, lms):
	(paths, mtracks) = mus.get_alltracks_db()
	mtrackids = mus.add_tracks( mtracks)
	mta = musly.MuslyTracksAdded( paths, mtracks, mtrackids)
	lsmworker = LMSWorker( lms, mus, mta)
	lsmworker.start()
	lsmworker.unpause()
	time.sleep(60*10)
	lsmworker.control("exit")
	time.sleep(2)

def test_webui( mus):
	from lmsmusly.web import webui
	_LOGGER.debug("webui enter")
	wu = webui.WebUI()
	cp = wu.start_server()
	_LOGGER.debug("webui exit")
	time.sleep(30)
	cp.engine.stop()
	cp.engine.exit()

def handle_stop_signals(signum, frame):
	global main_run
	main_run = False
	_LOGGER.warn( "received signal #{}, exiting".format( signum))

def handle_webui_exit():
	global main_run
	main_run = False
	_LOGGER.info( "received quit from WebUI, exiting")

def prog_atexit( pidfile):
	if pidfile:
		os.unlink( pidfile)

def analyze_music( mus, path, rep):
	if os.path.isfile(path):
		allfiles = tools.get_files_inputfile( path)
	elif os.path.isdir(path):
		allfiles = tools.get_files_dir( path)
	else:
		_LOGGER.error( "analyze_music: {} not found".format(path))
		return
	if allfiles:
		mtracks = mus.analyze_files( allfiles, rep)
#		_LOGGER.info( "analyze_files done: {} files".format( len( mtracks)))

def init_muslytracksadded( mus, read_jukebox):
	(paths, mtracks) = mus.get_alltracks_db()
	if not read_jukebox:
		mtrackids = mus.add_tracks( mtracks)
	else:
		mtrackids = mus.get_jukebox_from_file()
		_LOGGER.debug( "init_muslytracksadded get_jukebox_from_file: numtracks {}/{}/{}".format( len( paths), len(mtracks), len( mtrackids) if mtrackids != None else "None"))
		if mtrackids == None:
			return None
		if len(mtracks) != len( mtrackids):
			_LOGGER.error( "init_muslytracksadded: number of tracks in database ({}) does not match jukebox ({}) -> re-create the jukebox".format( len(mtracks), len( mtrackids)))
			return None
	return musly.MuslyTracksAdded( paths, mtracks, mtrackids)

def static_playlist( mus, mta, lms, numtracks, seedtrack_index = 0, seedtrack_file = None):
	lms.update()
	_LOGGER.info( "Server info: {}".format( lms))
	if not lms.is_player_known( config.LMS_PLAYER):
		_LOGGER.error( "static_playlist: player: {} not found".format( config.LMS_PLAYER))
		return
	if seedtrack_file:
		seedtrack = seedtrack_file
	else:
		seedtrack = lms.player( config.LMS_PLAYER).get_track_fileurl( seedtrack_index)
		if seedtrack is None:
			_LOGGER.error("static_playlist: track #{} not found in playlist".format( seedtrack_index))
			return

#	mta = init_muslytracksadded( mus, read_jukebox)
#	if mta == None:
#		return

	try:
		sid = mta.paths.index( seedtrack)
	except ValueError:
		_LOGGER.error("static_playlist: track {} not found in database".format( seedtrack))
		return
	_LOGGER.info( "static_playlist: building playlist based on track {}".format( seedtrack))
	( rt, rsims) = mus.get_similars( mta.mtracks, mta.mtrackids, sid, numtracks+1)
	for i in range( 1,len(rt)):
		_LOGGER.debug( "static_playlist result id: {:3} sim: {:8.6f} file: {}".format( rt[i], rsims[i], mta.paths[ rt[i]]))
		lms.player( config.LMS_PLAYER).playlist_add( mta.paths[ rt[i]])
	_LOGGER.info( "player {}: {} tracks added".format(config.LMS_PLAYER, len(rt)-1))
	lms.player( config.LMS_PLAYER).display( "{} tracks added".format(len(rt)-1))

def dynamic_playlist( mus, mta, lms, cherrypy_server, numtracks, seedtrackid, start_paused = False):
	config.LMS_SEEDTRACKIDX = seedtrackid
	config.LMS_MINPLSLEN = numtracks

#	mta = init_muslytracksadded( mus, read_jukebox)
#	if mta == None:
#		return
	lsmworker = LMSWorker( lms, mus, mta)
	if cherrypy_server:
		cherrypy_server.engine.subscribe("lm-exit",handle_webui_exit)
		cherrypy_server.engine.subscribe("lm-pause", lsmworker.pause)
		cherrypy_server.engine.subscribe("lm-unpause", lsmworker.unpause)
	lsmworker.start()
	if not start_paused:
		lsmworker.unpause()
	while main_run:
		time.sleep( 5)
	lsmworker.unpause()
	lsmworker.control("exit")
	lsmworker.join()

def show_stats( mus, scursor):
	_LOGGER.info("LMSmusly v{} from {}".format( __version__, __updated__))
	stats = mus.get_stats()
	_LOGGER.info("libmusly version: {} ".format( stats['musly_version']))
	_LOGGER.info("libmusly method information: {} ".format( stats['aboutmethod']))

	scursor.execute('SELECT file FROM tracks')
	paths = []
	for row in scursor:
		paths.append( row[0])
	_LOGGER.info("tracks in database: {} ".format( len(paths)))
	if len(paths) > 0:
		_LOGGER.info("common path of tracks: {} ".format( os.path.commonpath( paths)))

def set_stats_webui( mus, scursor):
	mus.set_stat("app_version","LMSmusly v{} from {}".format( __version__, __updated__))
	scursor.execute('SELECT count(vals) FROM tracks')
	mus.set_stat( 'numtracksdb', scursor.fetchone()[0])
	mus.set_stat( 'dbpath', config.MUSLY_DBPATH)
	mus.set_stat( 'plslen', config.LMS_MINPLSLEN)
	mus.set_stat( 'seedtrackidx', config.LMS_SEEDTRACKIDX)
	mus.set_stat( 'topresfact', config.MTOPRETFACT)
	if config.MSKIPDUPS:
		mus.set_stat( 'skipdupslen', config.MSKIPDUPS_LEN)

def write_jukebox( mus):
	_LOGGER.info("reading database")
	(paths, mtracks) = mus.get_alltracks_db()
	mtrackids = mus.add_tracks( mtracks)
	_LOGGER.info("created jukebox with {} tracks".format( mus.get_numtracks()))
	mus.write_jukebox()
	_LOGGER.info("jukebox written to {}".format( os.path.join(config.MUSLY_DBPATH,config.MUSLY_JUKEBOXNAME)))

def main(argv=None): # IGNORE:C0111
	'''Command line options.'''

	if argv is None:
		argv = sys.argv
	else:
		sys.argv.extend(argv)

	program_name = os.path.basename(sys.argv[0])
	program_version = "v%s" % __version__
	program_build_date = str(__updated__)
	program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
	program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
	program_license = '''%s

	LMSmusly v%s.
	Copyright 2017 R.S.U. GPL v3. All rights reserved.
	https://www.nexus0.net/pub/sw/lmsmusly

USAGE
''' % (program_shortdesc, str(__version__))

	try:
		# Setup argument parser
		parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
		parser.add_argument('-V', '--version', action='version', version=program_version_message)
		parser.add_argument('--log-level', action='store', choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'], default='INFO', help="set log level (default: %(default)s)")

		parser.add_argument('--db-path', action='store', required=True, help='Database path')
#		parser.add_argument('--musly-lib', action='store', default=config.MUSLY_LIB, help='Path to libmusly.so (default: %(default)s)')
		parser.add_argument('--pid-file', action='store', help='write PID to a file')

		parent_parser = ArgumentParser(add_help=False)
		parent_parser.add_argument('--lms', action='store', required=True, help='LMS server name')
		parent_parser.add_argument('--lms-port', action='store', type=int, default=9000, help='LMS server port (default: %(default)s)')
		parent_parser.add_argument('--lms-player', action='store', required=True, help='LMS player ID. Format: 00:00:00:00:00:00')
		parent_parser.add_argument('--lms-notifications', action='store_true', default=False, help='show notifications on player')
		parent_parser.add_argument('--read-jukebox', action='store_true', default=False, help='read on-disk jukebox on startup')

		subparsers = parser.add_subparsers(title='operation modes', dest='subparser_name')
		analyze_parser = subparsers.add_parser('analyze', help='analyze music')
		static_parser = subparsers.add_parser('static', parents=[parent_parser], help='create static playlist')
		dynamic_parser = subparsers.add_parser('dynamic', parents=[parent_parser], help='serve dynamic playlist')
		stats_parser = subparsers.add_parser('info', help='print information')
		jukebox_parser = subparsers.add_parser('writejukebox', help='create on-disk jukebox')

		analyze_parser.add_argument('input', action='store', help='Directory to analyze or file to read files to analyze from')
		analyze_parser.add_argument('--replace-path', action='store',help='replace pathname. Format: old_path:new_path')
		analyze_parser.add_argument('--extract-length', action='store', type=int, default=config.MUSLY_EXTRACTLEN, help='extraction length (default: %(default)s sec)')
		analyze_parser.add_argument('--extract-start', action='store', type=int, default=config.MUSLY_EXTRACTSTART, help='extraction start (default: %(default)s sec)')

		static_parser.add_argument('numtracks', action='store', type=int, default=50, help='Number of tracks to add to playlist')
		spgroup = static_parser.add_mutually_exclusive_group()
		spgroup.add_argument('--seedtrack-index', action='store', type=int, default=0, help='playlist index of track to use (default: %(default)s)')
		spgroup.add_argument('--seedtrack-file', action='store', help='file name of track to use')

		dynamic_parser.add_argument('--no-duplicates', action='store', type=int, nargs='?', const=config.MSKIPDUPS_LEN, metavar="N", help='Avoid repeating a track for the next N tracks (default: %(const)s)')
		dynamic_parser.add_argument('--numtracks', action='store', type=int, default=10, help='target length of playlist (default: %(default)s)')
		dynamic_parser.add_argument('--seedtrack-index', action='store', type=int, default=config.LMS_SEEDTRACKIDX, help='playlist index of track to use as seed (default: %(default)s)')
		dynamic_parser.add_argument('--topresults-factor', action='store', type=int, default=config.MTOPRETFACT, help='top results selection factor (default: %(default)s)')
		dynamic_parser.add_argument('--webui', action='store_true', default=False, help='provide a web-based user interface')
		dynamic_parser.add_argument('--webui-host', action='store', default=config.WEBUI_HOST, help='host name / IP address for WebUI (default: %(default)s)')
		dynamic_parser.add_argument('--webui-port', action='store', type=int, default=config.WEBUI_PORT, help='port for WebUI (default: %(default)s)')
		dynamic_parser.add_argument('--webui-logpath', action='store', default=config.WEBUI_LOGPATH, help='logfile path for WebUI (default: %(default)s)')
		dynamic_parser.add_argument('--start-paused', action='store_true', default=False, help='pause the dynamic playlist on startup')

		# Process arguments
		args = parser.parse_args()
#		if args.musly_lib:
#			config.MUSLY_LIB = os.path.abspath( args.musly_lib)
		if args.db_path:
			config.MUSLY_DBPATH = os.path.abspath( args.db_path)
		if args.pid_file:
			with open(args.pid_file, 'w') as f:
				atexit.register( prog_atexit, pidfile = args.pid_file)
				f.write( str(os.getpid()))
		
		logging.basicConfig(level=args.log_level)

#		_LOGGER.info("libmusly default: {} ".format( config.MUSLY_LIB))
		_LOGGER.info("database: {} ".format( os.path.join(config.MUSLY_DBPATH,config.MUSLY_DBNAME)))

		sconn = sqlite3.connect( os.path.join(config.MUSLY_DBPATH,config.MUSLY_DBNAME))
		scursor = sconn.cursor()
		tools.db_init( scursor)

		mus = musly.Musly( scursor)

		global main_run
		main_run = True

		signal.signal(signal.SIGINT, handle_stop_signals)
		signal.signal(signal.SIGTERM, handle_stop_signals)
		
		cherrypy_server = None

		if args.subparser_name == 'analyze':
			_LOGGER.debug("analyze: input: {} rep: {}".format( args.input, args.replace_path))
			if args.extract_length:
				config.MUSLY_EXTRACTLEN = args.extract_length
			if args.extract_start:
				config.MUSLY_EXTRACTSTART = args.extract_start
			analyze_music( mus, args.input, args.replace_path)
			sconn.commit()
		elif args.subparser_name == 'info':
			show_stats( mus, scursor)
		elif args.subparser_name == 'static' or args.subparser_name == 'dynamic':
			if args.lms_notifications:
				config.LMS_NOTIFICATIONS = True
			if args.lms_port:
				config.LMS_SERVERPORT = args.lms_port
			config.LMS_SERVERNAME = args.lms
			config.LMS_PLAYER= args.lms_player
			_LOGGER.debug("LMS: name: {} port: {} player: {}".format( config.LMS_SERVERNAME, config.LMS_SERVERPORT, config.LMS_PLAYER))
			lms = LMSServer( config.LMS_SERVERNAME, config.LMS_SERVERPORT) 
			mta = init_muslytracksadded( mus, args.read_jukebox)

			if mta != None and args.subparser_name == 'static':
				_LOGGER.debug("static: numtracks: {} seedtrack_idx: {} seedtrack_file: {}".format( args.numtracks, args.seedtrack_index, args.seedtrack_file))
				static_playlist( mus, mta, lms, args.numtracks, args.seedtrack_index, args.seedtrack_file)
			elif mta != None and args.subparser_name == 'dynamic':
				if args.no_duplicates:
					_LOGGER.debug( "dynamic: avoid dups: {}".format( args.no_duplicates))
					config.MSKIPDUPS = True
					config.MSKIPDUPS_LEN = args.no_duplicates
				if args.seedtrack_index:
					config.LMS_SEEDTRACKIDX = args.seedtrack_index
				_LOGGER.debug("dynamic: numtracks: {} seedtrack: {}".format( args.numtracks, config.LMS_SEEDTRACKIDX))
				if args.webui:
					if args.webui_logpath:
						config.WEBUI_LOGPATH = args.webui_logpath
					if args.webui_port:
						config.WEBUI_PORT = args.webui_port
					if args.webui_host:
						config.WEBUI_HOST = args.webui_host
					if args.topresults_factor:
						config.MTOPRETFACT = args.topresults_factor
						config.MMAXRETFAC = config.MTOPRETFACT + 4
					from lmsmusly.web import webui
					wu = webui.WebUI( mus, mta, lms)
					wu.start_server()
					cherrypy_server = wu.get_cherrypy()
					set_stats_webui( mus, scursor)
				dynamic_playlist( mus, mta, lms, cherrypy_server, args.numtracks, config.LMS_SEEDTRACKIDX, args.start_paused)
		elif args.subparser_name == 'writejukebox':
			write_jukebox( mus)

		if args.subparser_name == 'dynamic' and args.webui and cherrypy_server:
			_LOGGER.debug("shutting down webserver")
			cherrypy_server.engine.stop()
			cherrypy_server.engine.exit()

		sconn.close()
		mus.jukebox_off()
		return 0
	except KeyboardInterrupt:
		### handle keyboard interrupt ###
		if mus:
			mus.jukebox_off()
		return 0
	except Exception as e:
		if DEBUG or TESTRUN:
			raise(e)
		indent = len(program_name) * " "
		sys.stderr.write(program_name + ": " + repr(e) + "\n")
		sys.stderr.write(indent + "  for help use --help")
		return 2

if __name__ == "__main__":
	sys.exit(main())
